@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 1997-2001, Michael Jennings
+ * Copyright (C) 1997-2002, Michael Jennings
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to
@@ -35,8 +35,8 @@
 # ifndef _BSD_SOURCE
 #  define _BSD_SOURCE
 # endif
-# ifndef inline
-#  define inline __inline__
+# ifndef _XOPEN_SOURCE
+#  define _XOPEN_SOURCE
 # endif
 #endif
 
@@ -44,7 +44,6 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <time.h>
-#include <sys/types.h>
 #include <sys/stat.h>
 #include <unistd.h>
 #include <ctype.h>
@@ -53,14 +52,23 @@
 #include <dirent.h>
 #include <errno.h>
 #include <signal.h>
+#ifdef TIME_WITH_SYS_TIME
+# include <sys/time.h>
+#endif
 #ifdef WITH_DMALLOC
 # include <dmalloc.h>
 #elif defined(HAVE_MALLOC_H)
 # include <malloc.h>
 #endif
-#ifdef HAVE_REGEX_H
-# include <regex.h>
-#endif
+
+#include <netdb.h>
+#include <sys/types.h>
+#include <sys/socket.h>
+#include <netinet/in.h>
+#include <sys/un.h>
+#include <arpa/inet.h>
+#include <netinet/tcp.h>
+#include <netinet/udp.h>
 
 #ifdef LIBAST_X11_SUPPORT
 # include <X11/Xatom.h>
@@ -71,12 +79,36 @@
 # endif
 #endif
 
-/******************************* GENERIC GOOP *********************************/
-#ifndef TRUE
-# define TRUE    ((unsigned char)(1))
-# define FALSE   ((unsigned char)(0))
+#ifdef LIBAST_REGEXP_SUPPORT_PCRE
+#  ifdef HAVE_PCRE_H
+#    include <pcre.h>
+#  elif HAVE_PCRE_PCRE_H
+#    include <pcre/pcre.h>
+#  endif
+#elif defined(LIBAST_REGEXP_SUPPORT_POSIX) || defined(LIBAST_REGEXP_SUPPORT_BSD)
+#  ifdef HAVE_REGEX_H
+#    include <regex.h>
+#  endif
+#  ifdef LIBAST_REGEXP_SUPPORT_BSD
+extern char *re_comp();
+extern int re_exec();
+#  endif
 #endif
 
+#include <libast/types.h>
+#include <libast/obj.h>
+#include <libast/regexp.h>
+#include <libast/socket.h>
+#include <libast/str.h>
+#include <libast/tok.h>
+#include <libast/url.h>
+
+#include <libast/list_if.h>
+#include <libast/array.h>
+#include <libast/linked_list.h>
+#include <libast/dlinked_list.h>
+
+/******************************* GENERIC GOOP *********************************/
 #define USE_VAR(x)   (void) x
 
 #ifdef MIN
@@ -89,27 +121,21 @@
 # define MIN(a,b)                       __extension__ ({__typeof__(a) aa = (a); __typeof__(b) bb = (b); (aa < bb) ? (aa) : (bb);})
 # define MAX(a,b)                       __extension__ ({__typeof__(a) aa = (a); __typeof__(b) bb = (b); (aa > bb) ? (aa) : (bb);})
 # define LOWER_BOUND(current, other)    __extension__ ({__typeof__(other) o = (other); ((current) < o) ? ((current) = o) : (current);})
-# define AT_LEAST(current, other)       LOWER_BOUND(current, other)
-# define MAX_IT(current, other)         LOWER_BOUND(current, other)
 # define UPPER_BOUND(current, other)    __extension__ ({__typeof__(other) o = (other); ((current) > o) ? ((current) = o) : (current);})
-# define AT_MOST(current, other)        UPPER_BOUND(current, other)
-# define MIN_IT(current, other)         UPPER_BOUND(current, other)
 # define BOUND(val, min, max)           __extension__ ({__typeof__(min) m1 = (min); __typeof__(max) m2 = (max); ((val) < m1) ? ((val) = m1) : (((val) > m2) ? ((val) = m2) : (val));})
-# define CONTAIN(val, min, max)         BOUND(val, min, max)
-# define SWAP_IT(one, two, tmp)         do {(tmp) = (one); (one) = (two); (two) = (tmp);} while (0)
 #else
 # define MIN(a,b)	                (((a) < (b)) ? (a) : (b))
 # define MAX(a,b)                       (((a) > (b)) ? (a) : (b))
 # define LOWER_BOUND(current, other)    (((current) < (other)) ? ((current) = (other)) : (current))
-# define AT_LEAST(current, other)       LOWER_BOUND(current, other)
-# define MAX_IT(current, other)         LOWER_BOUND(current, other)
 # define UPPER_BOUND(current, other)    (((current) > (other)) ? ((current) = (other)) : (current))
-# define AT_MOST(current, other)        UPPER_BOUND(current, other)
-# define MIN_IT(current, other)         UPPER_BOUND(current, other)
 # define BOUND(val, min, max)           (((val) < (min)) ? ((val) = (min)) : (((val) > (max)) ? ((val) = (max)) : (val)))
-# define CONTAIN(val, min, max)         BOUND(val, min, max)
-# define SWAP_IT(one, two, tmp)         do {(tmp) = (one); (one) = (two); (two) = (tmp);} while (0)
 #endif
+#define AT_LEAST(current, other)        LOWER_BOUND(current, other)
+#define MAX_IT(current, other)          LOWER_BOUND(current, other)
+#define AT_MOST(current, other)         UPPER_BOUND(current, other)
+#define MIN_IT(current, other)          UPPER_BOUND(current, other)
+#define CONTAIN(val, min, max)          BOUND(val, min, max)
+#define SWAP_IT(one, two, tmp)          do {(tmp) = (one); (one) = (two); (two) = (tmp);} while (0)
 
 /****************************** DEBUGGING GOOP ********************************/
 #ifndef LIBAST_DEBUG_FD
@@ -146,12 +172,13 @@
 /* A quick and dirty macro to say, "Hi!  I got here without crashing!" */
 #define MOO()  do {__DEBUG(); libast_dprintf("Moo.\n");} while (0)
 
-/* Assertion/abort macros which are quite a bit more useful than assert() and abort(). */
+/* Assertion/abort macros which are quite a bit more useful than assert() and abort().  These are defined
+   differently based on the use of gcc, the presence of __FILE__ and __LINE__, and the DEBUG setting. */
 #if DEBUG >= 1
 # if defined(__FILE__) && defined(__LINE__)
 #  ifdef __GNUC__
 #   define ASSERT(x)  do {if (!(x)) {if (DEBUG_LEVEL>=1) {fatal_error("ASSERT failed in %s() at %s:%d:  %s\n", __FUNCTION__, __FILE__, __LINE__, #x);} \
-                                                    else {print_warning("ASSERT failed in %s() at %s:%d:  %s\n", __FUNCTION__, __FILE__, __LINE__, #x);}}} while (0)
+                                                    else {print_warning("ASSERT failed in %s() at %s:%d:  %s\n", __FUNCTION__, __FILE__, __LINE__, #x); return;}}} while (0)
 #   define ASSERT_RVAL(x, val)  do {if (!(x)) {if (DEBUG_LEVEL>=1) {fatal_error("ASSERT failed in %s() at %s:%d:  %s\n", __FUNCTION__, __FILE__, __LINE__, #x);} \
                                                               else {print_warning("ASSERT failed in %s() at %s:%d:  %s\n", __FUNCTION__, __FILE__, __LINE__, #x);} \
                                                return (val);}} while (0)
@@ -164,7 +191,7 @@
 #   define ABORT() fatal_error("Aborting in %s() at %s:%d.\n", __FUNCTION__, __FILE__, __LINE__)
 #  else
 #   define ASSERT(x)  do {if (!(x)) {if (DEBUG_LEVEL>=1) {fatal_error("ASSERT failed at %s:%d:  %s\n", __FILE__, __LINE__, #x);} \
-                                                    else {print_warning("ASSERT failed at %s:%d:  %s\n", __FILE__, __LINE__, #x);}}} while (0)
+                                                    else {print_warning("ASSERT failed at %s:%d:  %s\n", __FILE__, __LINE__, #x); return;}}} while (0)
 #   define ASSERT_RVAL(x, val)  do {if (!(x)) {if (DEBUG_LEVEL>=1) {fatal_error("ASSERT failed at %s:%d:  %s\n", __FILE__, __LINE__, #x);} \
                                                               else {print_warning("ASSERT failed at %s:%d:  %s\n", __FILE__, __LINE__, #x);} \
                                                return (val);}} while (0)
@@ -178,7 +205,7 @@
 #  endif
 # else
 #  define ASSERT(x)  do {if (!(x)) {if (DEBUG_LEVEL>=1) {fatal_error("ASSERT failed:  %s\n", #x);} \
-                                                   else {print_warning("ASSERT failed:  %s\n", #x);}}} while (0)
+                                                   else {print_warning("ASSERT failed:  %s\n", #x); return;}}} while (0)
 #  define ASSERT_RVAL(x, val)  do {if (!(x)) {if (DEBUG_LEVEL>=1) {fatal_error("ASSERT failed:  %s\n", #x);} \
                                                              else {print_warning("ASSERT failed:  %s\n", #x);} return (val);}} while (0)
 #  define ASSERT_NOTREACHED()           return
@@ -197,7 +224,7 @@
 # define REQUIRE_RVAL(x, v)             do {if (!(x)) return (v);} while (0)
 #endif
 
-#define NONULL(x) ((x) ? (x) : ("<null>"))
+#define NONULL(x) (((char *) (x)) ? ((char *) (x)) : ((char *) ("<" #x " null>")))
 
 /* Macros for printing debugging messages */
 #if DEBUG >= 1
@@ -231,6 +258,10 @@
 /* Use this for stuff that you only want turned on in dire situations */
 #define D_NEVER(x)             NOP
 
+#define DEBUG_OPTIONS          1
+#define D_OPTIONS(x)           DPRINTF1(x)
+#define DEBUG_OBJ              2
+#define D_OBJ(x)               DPRINTF2(x)
 #define DEBUG_CONF             3
 #define D_CONF(x)              DPRINTF3(x)
 #define DEBUG_MEM              5
@@ -291,9 +322,9 @@
 
 /* Fast memset() macro contributed by vendu */
 #if (SIZEOF_LONG == 8)
-# define MEMSET_LONG() l |= l<<32
+# define MEMSET_LONG() (l |= l<<32)
 #else
-# define MEMSET_LONG() ((void)0)
+# define MEMSET_LONG() NOP
 #endif
 
 #define MEMSET(s, c, count) do { \
@@ -339,10 +370,11 @@
 /******************************* STRINGS GOOP *********************************/
 
 #ifdef __GNUC__
-# define SWAP(a, b)  __extension__ ({__typeof__(a) tmp = (a); (a) = (b); (b) = tmp;})
+# define SWAP(a, b)  (void) __extension__ ({__typeof__(a) tmp = (a); (a) = (b); (b) = tmp;})
 #else
 # define SWAP(a, b)  do {void *tmp = ((void *)(a)); (a) = (b); (b) = tmp;} while (0)
 #endif
+#define BINSWAP(a, b)  (((long) (a)) ^= ((long) (b)) ^= ((long) (a)) ^= ((long) (b)))
 
 #define CONST_STRLEN(x)            (sizeof(x) - 1)
 #define BEG_STRCASECMP(s, constr)  (strncasecmp(s, constr, CONST_STRLEN(constr)))
@@ -358,6 +390,7 @@
 #  define PATH_MAX 255
 #endif
 
+#define CONFIG_BUFF                     20480
 #define CONF_BEGIN_CHAR                 ((char) 1)
 #define CONF_BEGIN_STRING               "\001"
 #define CONF_END_CHAR                   ((char) 2)
@@ -471,6 +504,108 @@ typedef struct conf_var_struct {
 
 extern fstate_t *fstate;
 extern unsigned char fstate_idx;
+extern const char *true_vals[], *false_vals[];
+
+
+/******************************* OPTIONS GOOP **********************************/
+
+/* Flags for individual options */
+#define SPIFOPT_FLAG_NONE                 (0)
+#define SPIFOPT_FLAG_BOOLEAN              (1UL << 0)
+#define SPIFOPT_FLAG_TYPEMASK_NOVALUE     (SPIFOPT_FLAG_BOOLEAN)
+#define SPIFOPT_FLAG_INTEGER              (1UL << 1)
+#define SPIFOPT_FLAG_STRING               (1UL << 2)
+#define SPIFOPT_FLAG_ARGLIST              (1UL << 3)
+#define SPIFOPT_FLAG_ABSTRACT             (1UL << 4)
+#define SPIFOPT_FLAG_TYPEMASK_VALUE       (SPIFOPT_FLAG_INTEGER | SPIFOPT_FLAG_STRING | SPIFOPT_FLAG_ARGLIST | SPIFOPT_FLAG_ABSTRACT)
+#define SPIFOPT_FLAG_TYPEMASK             (SPIFOPT_FLAG_TYPEMASK_NOVALUE | SPIFOPT_FLAG_TYPEMASK_VALUE)
+
+#define SPIFOPT_FLAG_PREPARSE             (1UL << 8)
+#define SPIFOPT_FLAG_DEPRECATED           (1UL << 9)
+
+/* Flags that control the parser's behavior */
+#define SPIFOPT_SETTING_POSTPARSE          (1UL << 0)
+
+#define SPIFOPT_OPTION(s, l, d, f, p, m)  { s, l, d,                                             (f), &(p), m }
+#define SPIFOPT_BOOL(s, l, d, v, m)       { s, l, d,                          (SPIFOPT_FLAG_BOOLEAN), &(v), m }
+#define SPIFOPT_BOOL_PP(s, l, d, v, m)    { s, l, d,  (SPIFOPT_FLAG_BOOLEAN | SPIFOPT_FLAG_PREPARSE), &(v), m }
+#define SPIFOPT_BOOL_LONG(l, d, v, m)     { 0, l, d,                          (SPIFOPT_FLAG_BOOLEAN), &(v), m }
+#define SPIFOPT_BOOL_LONG_PP(l, d, v, m)  { 0, l, d,  (SPIFOPT_FLAG_BOOLEAN | SPIFOPT_FLAG_PREPARSE), &(v), m }
+#define SPIFOPT_INT(s, l, d, p)           { s, l, d,                          (SPIFOPT_FLAG_INTEGER), &(p), 0 }
+#define SPIFOPT_INT_PP(s, l, d, p)        { s, l, d,  (SPIFOPT_FLAG_INTEGER | SPIFOPT_FLAG_PREPARSE), &(p), 0 }
+#define SPIFOPT_INT_LONG(l, d, p)         { 0, l, d,                          (SPIFOPT_FLAG_INTEGER), &(p), 0 }
+#define SPIFOPT_INT_LONG_PP(l, d, p)      { 0, l, d,  (SPIFOPT_FLAG_INTEGER | SPIFOPT_FLAG_PREPARSE), &(p), 0 }
+#define SPIFOPT_STR(s, l, d, p)           { s, l, d,                           (SPIFOPT_FLAG_STRING), &(p), 0 }
+#define SPIFOPT_STR_PP(s, l, d, p)        { s, l, d,   (SPIFOPT_FLAG_STRING | SPIFOPT_FLAG_PREPARSE), &(p), 0 }
+#define SPIFOPT_STR_LONG(l, d, p)         { 0, l, d,                           (SPIFOPT_FLAG_STRING), &(p), 0 }
+#define SPIFOPT_STR_LONG_PP(l, d, p)      { 0, l, d,   (SPIFOPT_FLAG_STRING | SPIFOPT_FLAG_PREPARSE), &(p), 0 }
+#define SPIFOPT_ARGS(s, l, d, p)          { s, l, d,                          (SPIFOPT_FLAG_ARGLIST), &(p), 0 }
+#define SPIFOPT_ARGS_PP(s, l, d, p)       { s, l, d,  (SPIFOPT_FLAG_ARGLIST | SPIFOPT_FLAG_PREPARSE), &(p), 0 }
+#define SPIFOPT_ARGS_LONG(l, d, p)        { 0, l, d,                          (SPIFOPT_FLAG_ARGLIST), &(p), 0 }
+#define SPIFOPT_ARGS_LONG_PP(l, d, p)     { 0, l, d,  (SPIFOPT_FLAG_ARGLIST | SPIFOPT_FLAG_PREPARSE), &(p), 0 }
+#define SPIFOPT_ABST(s, l, d, f)          { s, l, d,                         (SPIFOPT_FLAG_ABSTRACT),  (f), 0 }
+#define SPIFOPT_ABST_PP(s, l, d, f)       { s, l, d, (SPIFOPT_FLAG_ABSTRACT | SPIFOPT_FLAG_PREPARSE),  (f), 0 }
+#define SPIFOPT_ABST_LONG(l, d, f)        { 0, l, d,                         (SPIFOPT_FLAG_ABSTRACT),  (f), 0 }
+#define SPIFOPT_ABST_LONG_PP(l, d, f)     { 0, l, d, (SPIFOPT_FLAG_ABSTRACT | SPIFOPT_FLAG_PREPARSE),  (f), 0 }
+
+#define SPIFOPT_TYPE(opt)                 (((spifopt_t) (opt)).flags & SPIFOPT_FLAG_TYPEMASK)
+#define SPIFOPT_OPT_TYPE(n)               (SPIFOPT_OPT_FLAGS(n) & SPIFOPT_FLAG_TYPEMASK)
+#define SPIFOPT_OPT_IS_BOOLEAN(n)         (SPIFOPT_OPT_FLAGS(n) & SPIFOPT_FLAG_BOOLEAN)
+#define SPIFOPT_OPT_IS_STRING(n)          (SPIFOPT_OPT_FLAGS(n) & SPIFOPT_FLAG_STRING)
+#define SPIFOPT_OPT_IS_INTEGER(n)         (SPIFOPT_OPT_FLAGS(n) & SPIFOPT_FLAG_INTEGER)
+#define SPIFOPT_OPT_IS_ARGLIST(n)         (SPIFOPT_OPT_FLAGS(n) & SPIFOPT_FLAG_ARGLIST)
+#define SPIFOPT_OPT_IS_ABSTRACT(n)        (SPIFOPT_OPT_FLAGS(n) & SPIFOPT_FLAG_ABSTRACT)
+#define SPIFOPT_OPT_IS_PREPARSE(n)        (SPIFOPT_OPT_FLAGS(n) & SPIFOPT_FLAG_PREPARSE)
+#define SPIFOPT_OPT_IS_DEPRECATED(n)      (SPIFOPT_OPT_FLAGS(n) & SPIFOPT_FLAG_DEPRECATED)
+#define SPIFOPT_OPT_NEEDS_VALUE(n)        (SPIFOPT_OPT_FLAGS(n) & (SPIFOPT_FLAG_STRING | SPIFOPT_FLAG_INTEGER | SPIFOPT_FLAG_ARGLIST))
+
+#define SPIFOPT_OPT_SHORT(n)              (SPIFOPT_OPTLIST(n).short_opt)
+#define SPIFOPT_OPT_LONG(n)               (SPIFOPT_OPTLIST(n).long_opt)
+#define SPIFOPT_OPT_DESC(n)               (SPIFOPT_OPTLIST(n).desc)
+#define SPIFOPT_OPT_FLAGS(n)              (SPIFOPT_OPTLIST(n).flags)
+#define SPIFOPT_OPT_VALUE(n)              (SPIFOPT_OPTLIST(n).value)
+#define SPIFOPT_OPT_MASK(n)               (SPIFOPT_OPTLIST(n).mask)
+
+#define SPIFOPT_OPTLIST(n)                (spifopt_settings.opt_list[((n) < (spifopt_settings.num_opts) ? (n) : (0))])
+#define SPIFOPT_OPTLIST_SET(l)            (spifopt_settings.opt_list = ((spifopt_t *) (l)))
+#define SPIFOPT_NUMOPTS_GET()             (spifopt_settings.num_opts)
+#define SPIFOPT_NUMOPTS_SET(n)            (spifopt_settings.num_opts = (n))
+#define SPIFOPT_FLAGS_GET()               (spifopt_settings.flags)
+#define SPIFOPT_FLAGS_SET(m)              (spifopt_settings.flags |= (m))
+#define SPIFOPT_FLAGS_IS_SET(m)           (spifopt_settings.flags & (m))
+#define SPIFOPT_FLAGS_CLEAR(m)            (spifopt_settings.flags &= ~(m))
+#define SPIFOPT_BADOPTS_GET()             (spifopt_settings.bad_opts)
+#define SPIFOPT_BADOPTS_SET(n)            (spifopt_settings.bad_opts = (n))
+#define SPIFOPT_ALLOWBAD_GET()            (spifopt_settings.allow_bad)
+#define SPIFOPT_ALLOWBAD_SET(n)           (spifopt_settings.allow_bad = (n))
+#define SPIFOPT_INDENT_GET()              (spifopt_settings.indent)
+#define SPIFOPT_INDENT_SET(n)             (spifopt_settings.indent = (n))
+#define SPIFOPT_HELPHANDLER               ((spifopt_settings.help_handler) ? (spifopt_settings.help_handler) : (spifopt_usage))
+#define SPIFOPT_HELPHANDLER_SET(f)        (spifopt_settings.help_handler = (f))
+
+typedef void (*spifopt_helphandler_t)();
+typedef void (*spifopt_abstract_handler_t)(char *);
+
+typedef struct spifopt_t_struct {
+    spif_char_t short_opt;
+    spif_charptr_t long_opt;
+    spif_charptr_t desc;
+    spif_uint32_t flags;
+    void *value;
+    spif_uint32_t mask;
+} spifopt_t;
+
+typedef struct spifopt_settings_t_struct {
+    spifopt_t *opt_list;
+    spif_uint16_t num_opts;
+    spif_uint32_t flags;
+    spif_uint8_t bad_opts, allow_bad;
+    spif_uint8_t indent;
+    spifopt_helphandler_t help_handler;
+} spifopt_settings_t;
+
+extern spifopt_settings_t spifopt_settings;
+
 
 
 
@@ -507,7 +642,7 @@ extern GC libast_x_create_gc(const char *, unsigned long, Display *, Drawable, u
 extern void libast_x_free_gc(const char *, const char *, unsigned long, Display *, GC);
 extern void libast_dump_gc_tables(void);
 #endif
-extern void free_array(void **, size_t);
+extern void free_array(void *, size_t);
 
 /* file.c */
 extern int libast_temp_file(char *, size_t);
@@ -516,9 +651,9 @@ extern int libast_temp_file(char *, size_t);
 extern char *left_str(const char *, unsigned long);
 extern char *mid_str(const char *, unsigned long, unsigned long);
 extern char *right_str(const char *, unsigned long);
-#if defined(HAVE_REGEX_H)
-extern unsigned char regexp_match(const char *, const char *);
-extern unsigned char regexp_match_r(const char *str, const char *pattern, regex_t **rexp);
+#if defined(LIBAST_REGEXP_SUPPORT_POSIX) && defined(HAVE_REGEX_H)
+extern spif_bool_t regexp_match(const char *, const char *);
+extern spif_bool_t regexp_match_r(const char *str, const char *pattern, regex_t **rexp);
 #endif
 extern char **split(const char *, const char *);
 extern char **split_regexp(const char *, const char *);
@@ -550,9 +685,13 @@ extern char *garbage_collect(char *, size_t);
 extern char *file_garbage_collect(char *, size_t);
 extern char *condense_whitespace(char *);
 extern void hex_dump(void *, size_t);
+extern spif_cmp_t version_compare(const char *, const char *);
 #ifndef HAVE_MEMMEM
 extern void *memmem(const void *, size_t, const void *, size_t);
 #endif
+#ifndef HAVE_STRNLEN
+extern size_t strnlen(const char *, size_t);
+#endif
 #ifndef HAVE_USLEEP
 extern void usleep(unsigned long);
 #endif
@@ -574,4 +713,8 @@ extern FILE *open_config_file(char *name);
 extern void conf_parse_line(FILE *fp, char *buff);
 extern char *conf_parse(char *conf_name, const char *dir, const char *path);
 
+/* options.c */
+extern void spifopt_parse(int, char **);
+extern void spifopt_usage(void);
+
 #endif /* _LIBAST_H_ */
